/* $Id: DomHelper.java,v 1.1 2006/05/30 19:21:10 gsson Exp $ */

/*
 * Copyright (c) 2006 Henrik Gustafsson <henrik.gustafsson@fnord.se>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package se.fnord.domhelper;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

/**
 * A class for navigating and modifying DOM-trees.
 */
public class DomHelper {

	private static final int BEFORE_FIRST = -1;
	private static final int IN = 0;
	private static final int AFTER_LAST = 1;
	
	protected Document document;

	protected Element currentElement;
	protected Node parentNode;
	protected int state;
	
	/**
	 *
	 * @param document
	 *     The DOM document to navigate
	 */
	public DomHelper(Document document) {
		this.document = document;
		root();
	}

	/**
	 * Moves to the top level in the hierarchy. Normally, this level
	 * has only one element, the root element.
	 * <i>Think <code>cd /</code></i>
	 */
	public void root() {
		this.currentElement = null;
		this.parentNode = document;
		this.state = DomHelper.BEFORE_FIRST;
	}
	
	/**
	 * Enters the currently selected node.
	 * <i>Think <code>cd &lt;node&gt;</code></i>
	 * 
	 * @throws DomHelperException
	 *     if the operation cannot be completed.
	 */
	public void enter() throws DomHelperException {
		if (state != DomHelper.IN)
			throw new DomHelperSelectionException();
			
		parentNode = currentElement;
		state = DomHelper.BEFORE_FIRST;
		currentElement = null;

	}

	/**
	 * Moves up one step in the hierarchy.
	 * <i>Think <code>cd ..</code></i>
	 * 
	 * @throws DomHelperException
	 *     if the operation cannot be completed.
	 */
	public void leave() throws DomHelperException {

		if (!(parentNode instanceof Element))
			throw new DomHelperSelectionException();

		currentElement = (Element)parentNode;
		parentNode = parentNode.getParentNode();
		state = DomHelper.IN;
	}

	/**
	 * Moves to before the first element in the current level.
	 *
	 */
	public void begin() {
		currentElement = null;
		state = DomHelper.BEFORE_FIRST;
	}

	/**
	 * Moves to after the first element in the current level.
	 *
	 */
	public void end() {
		currentElement = null;
		state = DomHelper.AFTER_LAST;
	}

	/**
	 * Selects the next element in the current level.
	 * 
	 * @return
	 *     'true' if another entry exists. 'false' otherwise.
	 */
	public boolean next() {
		Node n;
		
		if (state == DomHelper.BEFORE_FIRST)
			n = parentNode.getFirstChild();
		else if (state == DomHelper.IN)
			n = currentElement.getNextSibling();
		else
			n = null;

		while (n != null && !(n instanceof Element))
			n = n.getNextSibling();
		
		if (n == null) {
			state = DomHelper.AFTER_LAST;
			currentElement = null;
			return false;
		}

		state = DomHelper.IN;
		currentElement = (Element)n;
		return true;
	}
	
	/**
	 * Selects the next element with the specified name in the current
	 * level.
	 * 
	 * @param name
	 *     name of element to look for.
	 * @return
	 *     'true' if another entry exists. 'false' otherwise.
	 */
	public boolean nextByTagName(String name) {
		Node n;
		
		if (state == DomHelper.BEFORE_FIRST)
			n = parentNode.getFirstChild();
		else if (state == DomHelper.IN)
			n = currentElement.getNextSibling();
		else
			n = null;
		
		while (n != null && (!(n instanceof Element) || !name.equals(n.getNodeName()))) {
			n = n.getNextSibling();
		}

		if (n == null) {
			state = DomHelper.AFTER_LAST;
			currentElement = null;
			return false;
		}
		
		state = DomHelper.IN;
		currentElement = (Element)n;
		return true;
	}

	/**
	 * Selects the previous element in the current level.
	 * 
	 * @return
	 *     'true' if another entry exists. 'false' otherwise.
	 */
	public boolean previous() {
		Node n;
		
		if (state == DomHelper.AFTER_LAST)
			n = parentNode.getLastChild();
		else if (state == DomHelper.IN)
			n = currentElement.getPreviousSibling();
		else
			n = null;

		while (n != null && !(n instanceof Element))
			n = n.getPreviousSibling();

		if (n == null) {
			state = DomHelper.BEFORE_FIRST;
			currentElement = null;
			return false;
		}

		state = DomHelper.IN;
		currentElement = (Element)n;
		return true;
	}

	/**
	 * Selects the previous element with the specified name in the
	 * current level.
	 * 
	 * @param name
	 *     name of element to look for.
	 * @return
	 *     'true' if another entry exists. 'false' otherwise.
	 */
	public boolean previousByTagName(String name) {
		Node n;
		
		if (state == DomHelper.AFTER_LAST)
			n = parentNode.getLastChild();
		else if (state == DomHelper.IN)
			n = currentElement.getPreviousSibling();
		else
			n = null;
		
		while (n != null && !(n instanceof Element) && !name.equals(n.getNodeName()))
			n = n.getPreviousSibling();

		if (n == null) {
			state = DomHelper.BEFORE_FIRST;
			currentElement = null;
			return false;
		}
		
		state = DomHelper.IN;
		currentElement = (Element)n;
		return true;
	}
	
	/**
	 * Inserts an element after the currently selected element in the
	 * current level and selects it.
	 * <i>Think <code>mkdir &lt;node&gt;</code></i>
	 * 
	 * @param e
	 *     element to insert.
	 */
	public void insert(Element e) {
		if (state == DomHelper.AFTER_LAST) {
			parentNode.appendChild(e);
		}
		else if (state == DomHelper.BEFORE_FIRST) {
			Node a = parentNode.getFirstChild();
			if (a == null)
				parentNode.appendChild(e);
			else
				parentNode.insertBefore(e, a);
		}
		else {
			Node a = currentElement.getNextSibling();
			if (a == null)
				parentNode.appendChild(e);
			else
				parentNode.insertBefore(e, a);
		}
		
		state = DomHelper.IN;
		currentElement = e;
	}
	
	/**
	 * Creates and inserts an element after the currently selected
	 * element in the current level and selects it.
 	 * <i>Think <code>mkdir &lt;node&gt;</code></i>
	 * 
	 * @param name
	 *     name of element to create and insert.
	 */
	public void insert(String name) {
		Element e = document.createElement(name);
		insert(e);
	}
	
	/**
	 * Removes the currently selected element and selects the previous
	 * element.
 	 * <i>Think <code>rmdir &lt;node&gt;</code></i>
	 * 
	 * @throws DomHelperException
	 *     if the operation cannot be completed.
	 */
	public void remove() throws DomHelperException {
		if (state != DomHelper.IN)
			throw new DomHelperSelectionException();
		
		Element e = currentElement;
		previous();
		parentNode.removeChild(e);
	}

	/**
	 * Removes all items in the current level.
	 * <i>Think <code>rmdir *</code></i>
	 * 
	 * @throws DomHelperException
	 *     if the operation cannot be completed.
	 */
	public void clear() throws DomHelperException {

		// If we are at root level, just remove the root element.
		if (parentNode instanceof Document) {
			Node n = parentNode.getFirstChild();
			if (n != null)
				parentNode.removeChild(n);
			return;
		}
		// Else, do a shallow clone of the parent element and replace
		// the current parent element with the new parent element.
		// TODO: Verify the presence/absence of attributes
		else if (parentNode instanceof Element ) {
				Node grandparentNode = parentNode.getParentNode();
				Element e = (Element)parentNode.cloneNode(false);
				grandparentNode.replaceChild(e, parentNode);
				parentNode = e;
				begin();
				return;
		}
		throw new DomHelperSelectionException();
	}
	
	/**
	 * Returns the text of the currently selected element.
	 * 
	 * @return
	 *     the name of all text-node children concatenated.
	 * @throws DomHelperException
	 *     if the operation could not be completed.
	 */
	public String getText() throws DomHelperException {
		if (state == DomHelper.IN) {
			StringBuffer buf = new StringBuffer();
			NodeList children = currentElement.getChildNodes();
	
			for (int i = 0; i < children.getLength(); i++) {
				Node child = children.item(i);
				
				if (child instanceof Text)
					buf.append(child.getNodeValue());
			}
			return buf.toString();
		}
		throw new DomHelperSelectionException();
	}

	/**
	 * Sets the text of the currently selected element.
	 * 
	 * @throws DomHelperException
	 *     if the operation could not be completed.
	 */
	public void setText(String text) throws DomHelperException {
		if (state == DomHelper.IN) {
			NodeList children = currentElement.getChildNodes();
	
			for (int i = 0; i < children.getLength(); i++) {
				Node child = children.item(i);
				
				if (child instanceof Text)
					currentElement.removeChild(child);
			}
			
			currentElement.appendChild(document.createTextNode(text));
		}
		else
			throw new DomHelperSelectionException();

	}

	/**
	 * Returns the name of the currently selected element.
	 * 
	 * @return
	 *     the name of the currently selected element.
	 * @throws DomHelperException
	 *     if the operation could not be completed.
	 */
	public String getNodeName() throws DomHelperException {
		if (state == DomHelper.IN)
			return currentElement.getNodeName();
		else
			throw new DomHelperSelectionException();
	}

	/**
	 * Returns the value of the named attribute of the currently
	 * selected element.
	 * 
	 * @param attribute
	 *     the name of the attribute whose value is to be returned.
	 * @return
	 *     the value of the attribute.
	 * @throws DomHelperException
	 *     if the operation could not be completed.
	 */
	public String getAttribute(String attribute) throws DomHelperException {
		if (state == DomHelper.IN)
			if (currentElement.hasAttribute(attribute))
				return currentElement.getAttribute(attribute);
		throw new DomHelperSelectionException();
	}

	/**
	 * Sets the value of the named attribute of the currently selected
	 * element.
	 * 
	 * @param attribute
	 *     the name of the attribute whose value is to be set.
	 * @param value
	 *     the value to set.
	 * @throws DomHelperException
	 *     if the operation could not be completed.
	 */
	public void setAttribute(String attribute, String value) throws DomHelperException {
		if (state == DomHelper.IN)
			currentElement.setAttribute(attribute, value);
		else
			throw new DomHelperSelectionException();
	}

	/**
	 * Tests whether the named attribute exists in the currently
	 * selected element.
	 * 
	 * @param attribute
	 *     the named of the attribute to test.
	 * @return
	 *     'true' if the attribute exists, 'false' otherwise.
	 * @throws DomHelperException
	 *     if the operation could not be completed.
	 */
	public boolean hasAttribute(String attribute) throws DomHelperException {
		if (state == DomHelper.IN)
			return currentElement.hasAttribute(attribute);
		throw new DomHelperSelectionException();
	}

	/**
	 * Returns the currently selected element.
	 * 
	 * @return
	 *     the currently selected element.
	 * @throws DomHelperException
	 *     if the operation could not be completed.
	 */
	public Element getElement() throws DomHelperException {
		if (state == DomHelper.IN)
			return currentElement;
		throw new DomHelperSelectionException();
	}
/*
//	private static final String SEPARATOR ="(//|/)";
//	private static final String ELEMENT = "(\\w+|\\*|\\.\\.|\\.)";		
	private static final String SEPARATOR ="(/)";
	private static final String ELEMENT = "(\\w+)";		
//	private static final Pattern path_pattern = Pattern.compile(SEPARATOR+ELEMENT+"(?:"+SEPARATOR+"$)?");

	private static final Pattern path_pattern = Pattern.compile("(?:^"+ELEMENT+")?"+SEPARATOR+ELEMENT+"(?:"+SEPARATOR+"$)?");
	
	class PathToken {
		public static final int DEFAULT_TOKEN = 0;
		public static final int SEPARATOR_TOKEN = 1;
		public static final int ELEMENT_TOKEN = 2;
		protected String token;
		PathToken(String token) {
			this.token = token;
		}
		
		int getType() {
			return PathToken.DEFAULT_TOKEN;
		}
		String getToken() {
			return token;
		}
	}
	
	class SeparatorToken extends PathToken{
		SeparatorToken(String token) {
			super(token);
		}
		
		int getType() {
			return PathToken.SEPARATOR_TOKEN;
		}
		String getToken() {
			return token;
		}
	}
	
	class ElementToken extends PathToken{
		ElementToken(String token) {
			super(token);
		}
		
		int getType() {
			return PathToken.ELEMENT_TOKEN;
		}
		String getToken() {
			return token;
		}
	}
	
	
	public void go(String path) throws DomHelperException {
		LinkedList s = new LinkedList();
		Matcher m = path_pattern.matcher(path);

		while(m.find()) {
			for (int i = 0; i < m.groupCount(); i++) {
				String token = m.group(i+1);
				if (token != null) {
					PathToken pt = null;
					switch(i) {
					case 0:
					case 2:
						pt = new ElementToken(token);
						break;
					case 1:
					case 3:
						pt = new SeparatorToken(token);
						break;
						
					}
					s.addFirst(pt);
				}
			}
		}
		
		PathToken t = (PathToken)s.getLast();
		if (t instanceof SeparatorToken)
			root();
		else
			begin();
		
		boolean deep = false;
		
		while (s.size() > 0) {
			t = (PathToken)s.removeLast();
			
			if (t instanceof SeparatorToken) {
				if (state == IN)
					enter();

				if (t.getToken().equals("//"))
					deep = true;
				else
					deep = false;
			}
			else {
				if (!nextByTagName(t.getToken()))
					throw new DomHelperSelectionException();
			}
		}
	}

	public static void main(String[] args) {
		
		try {
		    DocumentBuilderFactory factory = DocumentBuilderFactory
	        .newInstance();
		    DocumentBuilder builder = factory.newDocumentBuilder();
		    Document document = builder.newDocument();
		    
		    DomHelper helper = new DomHelper(document);

		    helper.root();
		    helper.clear();
		    helper.insert("kalle");
		    helper.enter();
		    helper.insert("pelle");
		    helper.insert("olle");
			helper.go("/kalle/olle");
			System.out.println("You are here: "+helper.getNodeName());
		} catch (DomHelperException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (ParserConfigurationException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
	}
*/
	
}
