/* $Id: ECalcView.java,v 1.22 2006/06/02 21:25:55 gsson Exp $ */

/*
 * Copyright (c) 2006 Henrik Gustafsson <henrik.gustafsson@fnord.se>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package se.fnord.ecalc.views;

import java.io.IOException;
import java.io.StringReader;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Properties;

import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.jface.action.ActionContributionItem;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.action.IContributionItem;
import org.eclipse.jface.action.IMenuListener;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.forms.widgets.FormText;
import org.eclipse.ui.forms.widgets.FormToolkit;
import org.eclipse.ui.forms.widgets.ScrolledForm;
import org.eclipse.ui.forms.widgets.ScrolledFormText;
import org.eclipse.ui.part.ViewPart;

import se.fnord.ecalc.Calculation;
import se.fnord.ecalc.ECalcPlugin;
import se.fnord.ecalc.StateStorage;
import se.fnord.ecalc.parser.EvaluationContext;
import se.fnord.ecalc.parser.EvaluationVisitor;
import se.fnord.ecalc.parser.ExpressionException;
import se.fnord.ecalc.parser.ExpressionParser;
import se.fnord.ecalc.parser.ExpressionParserVisitor;
import se.fnord.ecalc.parser.IOperators;
import se.fnord.ecalc.parser.IValue;
import se.fnord.ecalc.parser.Literal;
import se.fnord.ecalc.parser.ParseException;
import se.fnord.ecalc.parser.SimpleNode;
import se.fnord.ecalc.parser.Status;
import se.fnord.ecalc.parser.TokenMgrError;
import se.fnord.ecalc.parser.functions.ArcCosine;
import se.fnord.ecalc.parser.functions.ArcSine;
import se.fnord.ecalc.parser.functions.ArcTangent;
import se.fnord.ecalc.parser.functions.Ceiling;
import se.fnord.ecalc.parser.functions.Cosine;
import se.fnord.ecalc.parser.functions.Floor;
import se.fnord.ecalc.parser.functions.Hypotenuse;
import se.fnord.ecalc.parser.functions.Maximum;
import se.fnord.ecalc.parser.functions.Minimum;
import se.fnord.ecalc.parser.functions.Random;
import se.fnord.ecalc.parser.functions.Round;
import se.fnord.ecalc.parser.functions.Sine;
import se.fnord.ecalc.parser.functions.SquareRoot;
import se.fnord.ecalc.parser.functions.Tangent;

public class ECalcView extends ViewPart implements SelectionListener,
		KeyListener, ModifyListener, IMenuListener {
	
	
	public final static int RADIX_BIN = 2;
	public final static int RADIX_OCT = 8;
	public final static int RADIX_DEC = 10;
	public final static int RADIX_HEX = 16;

	public int displayRadix = RADIX_DEC;

	private List calculations = new LinkedList();
	private EvaluationContext context = new EvaluationContext();
	private Map displayRadixIdMap = new HashMap();
	private Map forceArithIdMap = new HashMap();
	private Map angleTypeIdMap = new HashMap();

	private ScrolledForm form;
	private List history = new LinkedList();
	private int history_position = -1;

	private Text input;

	private Properties messages = new Properties();
	
	private ScrolledFormText resultsContainer;
	private FormText resultsView;
	
	private FormToolkit toolkit;
	
	public ECalcView() {
		populateForceArithIdMap();
		populateDisplayTypeIdMap();
		populateAngleTypeIdMap();
		loadMessages();
		
		try {
			restoreState();

			context.setFunction("sqrt", new SquareRoot());
			context.setFunction("sin", new Sine());
			context.setFunction("cos", new Cosine());
			context.setFunction("tan", new Tangent());
			context.setFunction("asin", new ArcSine());
			context.setFunction("acos", new ArcCosine());
			context.setFunction("atan", new ArcTangent());
			context.setFunction("round", new Round());
			context.setFunction("ceil", new Ceiling());
			context.setFunction("floor", new Floor());
			context.setFunction("random", new Random());
			context.setFunction("hypot", new Hypotenuse());
			context.setFunction("max", new Maximum());
			context.setFunction("min", new Minimum());
			context.setConstant("pi", new Literal(Math.PI));
			context.setConstant("e", new Literal(Math.E));
		}
		catch (ExpressionException e) {
			e.printStackTrace();
		}
	}

	private void populateDisplayTypeIdMap() {
		displayRadixIdMap.put("se.fnord.ecalc.actions.displaybin", new Integer(RADIX_BIN));
		displayRadixIdMap.put("se.fnord.ecalc.actions.displayoct", new Integer(RADIX_OCT));
		displayRadixIdMap.put("se.fnord.ecalc.actions.displaydec", new Integer(RADIX_DEC));
		displayRadixIdMap.put("se.fnord.ecalc.actions.displayhex", new Integer(RADIX_HEX));
	}

	private void populateForceArithIdMap() {
		forceArithIdMap.put("se.fnord.ecalc.actions.forcetypeaction.doubletype", new Integer(IValue.TYPE_DOUBLE));
		forceArithIdMap.put("se.fnord.ecalc.actions.forcetypeaction.inttype", new Integer(IValue.TYPE_LONG));
		forceArithIdMap.put("se.fnord.ecalc.actions.forcetypeaction.autotype", new Integer(IValue.TYPE_AUTO));
	}

	private void populateAngleTypeIdMap() {
		angleTypeIdMap.put("se.fnord.ecalc.actions.angletype.rad", new Integer(IOperators.ANGLE_RAD));
		angleTypeIdMap.put("se.fnord.ecalc.actions.angletype.deg", new Integer(IOperators.ANGLE_DEG));
	}

	private void addCalculation(String expression, String result) {
		if (calculations.size() == 50) {
			calculations.remove(0);
		}
		calculations.add(new Calculation(expression, result));
		addHistory(expression);
	}

	private boolean addHistory(String expression) {
		if (expression.matches(".*\\S+.*")) {
			ListIterator i = history.listIterator();
			while (i.hasNext()) {
				String c = (String) i.next();
				if (c.equals(expression)) {
					i.remove();
					break;
				}
			}

			if (history.size() == 50) {
				history.remove(0);
			}
			history.add(expression);
			return true;
		}
		else {
			return false;
		}
	}

	private void buildResults() {
		StringBuffer s = new StringBuffer("<form>");
		ListIterator i = calculations.listIterator();
		while (i.hasNext()) {
			Calculation c = (Calculation) i.next();
			s.append(c.getFormatted(messages));
		}

		s.append("</form>");

		resultsContainer.setRedraw(false);
		resultsView.setText(s.toString(), true, false);
		resultsContainer.reflow(false);
		resultsContainer.reflow(false);
		resultsContainer.setOrigin(0, resultsView.getBounds().height);
		resultsContainer.setRedraw(true);
	}

	public void clearHistory() {
		calculations.clear();
		buildResults();
	}

	public void createPartControl(Composite parent) {
		toolkit = new FormToolkit(parent.getDisplay());
		form = toolkit.createScrolledForm(parent);
		form.setRedraw(false);
		GridLayout l = new GridLayout(1, true);
		form.getBody().setLayout(l);

		resultsContainer = new ScrolledFormText(form.getBody(), false);

		resultsView = toolkit.createFormText(resultsContainer, false);
		resultsView.setParagraphsSeparated(true);

		resultsContainer.setMinWidth(100);
		resultsContainer.setExpandHorizontal(true);
		resultsContainer.setMinHeight(100);
		resultsContainer.setExpandVertical(true);

		resultsContainer.setAlwaysShowScrollBars(false);
		resultsContainer.setFormText(resultsView);
		toolkit.adapt(resultsContainer);
		resultsContainer.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true,
				true));

		Label separator = toolkit.createSeparator(form.getBody(),
				SWT.HORIZONTAL);
		separator.setLayoutData(new GridData(SWT.FILL, SWT.BOTTOM, true, false));

		input = toolkit.createText(form.getBody(), "");
		input.setLayoutData(new GridData(SWT.FILL, SWT.BOTTOM, true, false));
		input.addSelectionListener(this);
		input.addKeyListener(this);
		input.addModifyListener(this);
		form.reflow(true);
		form.setRedraw(true);

		this.getViewSite().getActionBars().getMenuManager().addMenuListener(this);
		buildResults();
	}

	public void dispose() {
		storeState();
		this.getViewSite().getActionBars().getMenuManager().saveWidgetState();

		super.dispose();
	}

	public int getDisplayRadixMapping(String id) {
		Integer i = (Integer)displayRadixIdMap.get(id);
		if (i == null)
			throw new IllegalArgumentException();
		return i.intValue();
	}

	public int getForceArithMapping(String id) {
		Integer i = (Integer)forceArithIdMap.get(id);
		if (i == null)
			throw new IllegalArgumentException();
		return i.intValue();
	}

	public void keyPressed(KeyEvent e) {
		if (e.keyCode == SWT.ARROW_UP) {
			e.doit = false;

			if (history_position == -1) {
				addHistory(input.getText());
				history_position = history.size() - 1;
				setFromHistory();
			}
			else if (history.size() > 0 && history_position > 0) {
				history_position--;
				setFromHistory();
			}
		}
		else if (e.keyCode == SWT.ARROW_DOWN) {
			e.doit = false;
			if (history_position == -1) {
				addHistory(input.getText());
				input.setText("");
			}
			else if (history_position < history.size() - 1) {
				history_position++;
				setFromHistory();
			}
			else {
				addHistory(input.getText());
				input.setText("");
				history_position = -1;
			}
		}
	}

	public void keyReleased(KeyEvent e) {}

	private void loadMessages() {
		ECalcPlugin plugin = ECalcPlugin.getDefault();
		try {
			messages.load(plugin.openStream(new Path("messages.properties"),
					true));

		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public void menuAboutToShow(IMenuManager manager) {
		IContributionItem[] items = manager.getItems();

		for (int i = 0; i < items.length; i++) {
			String id = items[i].getId();
			Integer value;
			if (forceArithIdMap.containsKey(id)) {
				IAction a = ((ActionContributionItem)items[i]).getAction();
				value = (Integer)forceArithIdMap.get(id);
				a.setChecked(context.arithmeticType == value.intValue());
			}
			else if (displayRadixIdMap.containsKey(id)) {
				IAction a = ((ActionContributionItem)items[i]).getAction();
				value = (Integer)displayRadixIdMap.get(id);
				a.setChecked(displayRadix == value.intValue());				
			}
			else if (angleTypeIdMap.containsKey(id)) {
				IAction a = ((ActionContributionItem)items[i]).getAction();
				value = (Integer)angleTypeIdMap.get(id);
				a.setChecked(context.angleType == value.intValue());				
			}
		}
		manager.removeMenuListener(this);
	}

	public void modifyText(ModifyEvent e) {
		// TODO: Really ugly construct. There must be a more elegant way of
		// doing this.
		if (history_position != -1
				&& !input.getText().equals(history.get(history_position)))
			history_position = -1;
	}

	private void restoreState() {
		ECalcPlugin plugin = ECalcPlugin.getDefault();

		int displayRadix = ECalcPlugin.getDefault().getPreferenceStore().getInt("displayRadix");
		if (displayRadix == 0)
			setDisplayRadix(RADIX_DEC);
		else
			setDisplayRadix(displayRadix);
			
		int forceArith = ECalcPlugin.getDefault().getPreferenceStore().getInt("forceArith");
		if (forceArith == 0)
			setForceArith(IValue.TYPE_AUTO);
		else
			setForceArith(forceArith);
		
		int angleType = ECalcPlugin.getDefault().getPreferenceStore().getInt("angleType");
		if (angleType == 0)
			setAngleType(IOperators.ANGLE_RAD);
		else
			setAngleType(IOperators.ANGLE_DEG);

		IPath state_location = plugin.getStateLocation();

		StateStorage ss = new StateStorage(state_location.addTrailingSeparator().append("state.xml").toFile());
		context.setAnswer(ss.getAnswer());
		context.setFunctions(ss.getFunctions());
		context.setVariables(ss.getVariables());
		history = ss.getHistory();
		calculations = ss.getCalculations();
		history_position = -1;

	}

	public void setDisplayRadix(int radix) {
		displayRadix = radix;
		ECalcPlugin.getDefault().getPreferenceStore().setValue("displayRadix", displayRadix);
	}

	public void setFocus() {
		input.setFocus();
	}

	public void setForceArith(int type) {
		context.arithmeticType = type;
		ECalcPlugin.getDefault().getPreferenceStore().setValue("forceArith", type);
	}

	private void setFromHistory() {
		if (history_position >= 0) {
			String entry = (String) history.get(history_position);
			input.setText(entry);
			input.setSelection(entry.length() + 1);
		}
		else {
			input.setText("");
		}
	}

	private void storeState() {
		ECalcPlugin plugin = ECalcPlugin.getDefault();
		plugin.savePluginPreferences();

		IPath state_location = plugin.getStateLocation();

		StateStorage ss = new StateStorage();
		ss.setAnswer(context.getAnswer());
		ss.setCalculations(calculations);
		ss.setHistory(history);
		ss.setVariables(context.getVariables());
		ss.setFunctions(context.getFunctions());
		ss.save(state_location.addTrailingSeparator().append("state.xml").toFile());
	}

	public void widgetDefaultSelected(SelectionEvent e) {

		ExpressionParser parser = null;
		String expression = input.getText();
		String result;
		if (expression.matches(".*\\S+.*")) {
			try {
				parser = new ExpressionParser(new StringReader(expression));
				SimpleNode parseTree = parser.Start();
				ExpressionParserVisitor visitor = new EvaluationVisitor();
				IValue value = parseTree.accept(visitor, context);

				if (value.getPreferredType() == IValue.TYPE_STATUS) {
					result = ((Status)value).getMessage();
				}
				else {
					context.setAnswer(value);
					
					if (value.getPreferredType() == IValue.TYPE_LONG) {
						switch(displayRadix) {
						case RADIX_BIN:
							result = "0b" + Long.toBinaryString(value.toLong());
							break;
						case RADIX_OCT:
							result = "0o" + Long.toOctalString(value.toLong());
							break;
						case RADIX_DEC:
							result = Long.toString(value.toLong());
							break;
						case RADIX_HEX:
							result = "0x" + Long.toHexString(value.toLong());
							break;
						default:
							throw new IllegalArgumentException();
						}
					}
					else result = Double.toString(value.toDouble());
				}
			} catch (ExpressionException exception) {
				result = exception.getMessageId();
			} catch (ParseException exception) {
				result = "%se.fnord.ecalc.exceptions.parseexception";
			} catch (TokenMgrError exception) {
				result = "%se.fnord.ecalc.exceptions.parseexception";
			}

			addCalculation(expression, result);
			buildResults();
			history_position = history.size() - 1;
			if (history_position != -1) {
				setFromHistory();
				input.setSelection(0, expression.length());
			}
		}
	}

	public void widgetSelected(SelectionEvent e) {}

	public int getAngleTypeMapping(String id) {
		Integer i = (Integer)angleTypeIdMap .get(id);
		if (i == null)
			throw new IllegalArgumentException();
		return i.intValue();
	}

	public void setAngleType(int angleType) {
		context.angleType = angleType;
		ECalcPlugin.getDefault().getPreferenceStore().setValue("angleType", angleType);
	}

}